{ Copyright 1997-2003 by Frank Heckenbach <heckenb@mi.uni-erlangen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. }

{$gnu-pascal,I+}

unit Tools;

interface

uses GPC, Commentator_Pas;

const
  haInteger      = 1;
  haGMP          = 2;
  haValenceElim  = 3;
  haValenceBB    = 4;

  hmGroup        = 0;
  hmBetti        = 1;
  hmNTorsionRank = 2;
  hmMatrix       = 3;
  hmMatrixFile   = 4;

  { Maximum allowed homology dimension }
  MaximumDim = 256;

  { Number of steps after which progress messages are given.
    0 means no progress messages. }
  MsgStep = 1000;

  { Maximum allowed facet dimension }
  MaxN = 256;

type
  Dim0 = 0 .. MaximumDim;
  Dim1 = 1 .. MaximumDim;

  { Type of the point indices in faces and facets }
  TElementCount = Cardinal;

  { General counter type }
  TCount = Integer;

  { Sparse integer vector element }
  TIntVectorElement = record
    Index: TCount;
    Val: Integer
  end;

  { Sparse integer vector }
  PSparseIntVector = ^TSparseIntVector;
  TSparseIntVector (Size: TCount) = record
    Count: 0 .. Size;
    Elements: array [0 .. Size - 1] of TIntVectorElement
  end;

  n0 = 0 .. MaxN;
  n1 = 1 .. MaxN;

  { Type for a simplex }
  PPFace = ^PFace;
  PFace = ^TFace;
  TFace = array [0 .. MaxN] of TElementCount;

  { List of facets, used while reading them }
  PFacetList = ^TFacetList;
  TFacetList = record
    Next: PFacetList;
    Facet: PFace
  end;

var
  Commentator: TCommentator;

  ComplexFile: Text;

  MultiComplexFlag: Boolean = False;

  { Definition of the characters for parsing the facets }

  { Beginning and ending of a facet
    Nested Ch_Start/Ch_End pairs are handled correctly. }
  Ch_Start: Char = '[';
  Ch_End: Char = ']';

  { Separator of the points of a facets }
  Ch_Sep: Char = ',';

  { Beginning of a complex }
  Ch_Start_Compl1: Char = '=';
  Ch_Start_Compl2: Char = '[';

  { Ending of a complex }
  Ch_End_Compl: Char = ']';

const
  { Whitespace characters }
  Chs_Whitespace = [' ', #9 { TAB }, #13 { CR }, #10 { LF }];

  { Characters to be ignored within points }
  Chs_Ignore = Chs_Whitespace;

  { Characters to be ignored between facets }
  Chs_Between = Chs_Whitespace + [',', ';'];

  { Characters to be ignored between several complexes }
  Chs_Ignore_Compl = Chs_Whitespace + [';'];

  { Size of the hash table used to store the point descriptions }
  HashSize = $4000;

  { Maximum allowed length of point descriptions }
  MaxLength = $ff;

{ Abort with an error message }
procedure Abort (const Message: String);

{ Read a characeter from the input }
function ReadChar: Char;

{ Parse facets from the input }
function ParseFacets (const Description: String; Sort: Boolean; var FacetCount, PointCount: TCount): PFacetList;

implementation

var
  LineRead: TString = '';
  LineReadIndex: Integer = 1;

procedure Abort (const Message: String);
begin
  ActivityReport (Commentator, Message, LvlNormal, Internal_Error);
  Halt (5)
end;

function ReadChar: Char;
begin
  while LineReadIndex > Length (LineRead) do
    if EOF (ComplexFile) then
      begin
        ReadChar := #0;
        Exit
      end
    else if EOLn (ComplexFile) then
      ReadLn (ComplexFile)
    else
      begin
        Read (ComplexFile, LineRead);
        LineRead := Trim (LineRead);
        LineReadIndex := 1
      end;
  ReadChar := LineRead[LineReadIndex];
  Inc (LineReadIndex)
end;

function ParseFacets (const Description: String; Sort: Boolean; var FacetCount, PointCount: TCount): PFacetList;
type
  { List of strings }
  PHashList = ^THashList;
  THashList = record
    Next: PHashList;
    Index: TElementCount;
    s: PString
  end;

var
  HashList: array [0 .. HashSize - 1] of PHashList;
  MaxIndex: TElementCount;

  procedure InitHash;
  var c: TCount;
  begin
    for c := 0 to HashSize - 1 do HashList[c] := nil;
    MaxIndex := 0
  end;

  procedure DoneHash;
  var
    c: TCount;
    ph: PHashList;
  begin
    for c := 0 to HashSize - 1 do
      while HashList[c] <> nil do
        begin
          ph := HashList[c];
          HashList[c] := HashList[c]^.Next;
          Dispose (ph)
        end
  end;

  { Search a given point description in the list of descriptions
    seen so far and return its index. If it did not occur yet, make
    a new index and return it. The indices made are contiguous and
    start with 1. }
  function GetIndex (const s: TString): TElementCount;
  var
    p: ^PHashList;
    i, n, Hash: Integer;
  begin
    Hash := Length (s);
    n := 10;
    for i := 1 to Length (s) do
      begin
        Hash := Hash + n * Ord (s[i]);
        n := (n * 27) mod HashSize
      end;
    Hash := Hash mod HashSize;
    p := @HashList[Hash];
    while (p^ <> nil) and (s <> p^^.s^) do p := @p^^.Next;
    if p^ = nil then
      begin
        New (p^);
        p^^.Next := nil;
        New (p^^.s, Length (s));
        p^^.s^ := s;
        Inc (MaxIndex);
        p^^.Index := MaxIndex
      end;
    GetIndex := p^^.Index
  end;

var
  SCount: TCount;
  FacetList, pf: PFacetList;
  Facet: TFace;
  i, j, z1, z2: n0;
  t: TElementCount;
  s: TString;
  ch: Char;
  f: Boolean;

begin
  InitHash;
  FacetList := nil;
  FacetCount := 0;

  { Read and parse the facets }
  repeat
    repeat
      ch := ReadChar
    until not (ch in Chs_Between);
    if (ch = #0) or (ch = Ch_End_Compl) then Break;

    { Beginning of a new facet }
    if ch <> Ch_Start then Abort ('Beginning of ' + Description + ' not found.');
    i := 0;

    { Read the points }
    repeat
      s := '';
      SCount := 0;
      repeat
        ch := ReadChar;
        if ch = #0 then Abort ('Unexpected end of file!');
        f := (SCount = 0) and ((ch = Ch_Sep) or (ch = Ch_End));
        if ch = Ch_Start then Inc (SCount);
        if ch = Ch_End then Dec (SCount);
        if not f and not (ch in Chs_Ignore) and (Length (s) < MaxLength) then
          Insert (ch + '', s, Length (s) + 1)
      until f;
      if Length (s) = MaxLength then Abort ('String too long (`' + s + ''')!');
      if i >= MaxN then Abort (Description + ' too big (increase MaxN)!');
      Facet[i] := GetIndex (s);
      Inc (i)
    until ch = Ch_End;

    { Sort the points using bubble sort. Efficiency does not matter here. }
    if Sort then
      begin
        if i >= 2 then
          for z1 := 0 to i - 2 do
            for z2 := i - 1 downto z1 + 1 do
              if Facet[z1] > Facet[z2] then
                begin
                  t := Facet[z1];
                  Facet[z1] := Facet[z2];
                  Facet[z2] := t
                end;

        { Remove duplicate points }
        z2 := 1;
        for z1 := 1 to i - 1 do
          if Facet[z1] <> Facet[z1 - 1] then
            begin
              Facet[z2] := Facet[z1];
              Inc (z2)
            end;
        i := z2
      end;
    Facet[i] := 0;

    { Store facet }
    pf := FacetList;
    New (FacetList);
    FacetList^.Next := pf;
    GetMem (FacetList^.Facet, (i + 1) * Sizeof (FacetList^.Facet^[0]));
    for j := 0 to i do FacetList^.Facet^[j] := Facet[j];
    Inc (FacetCount)

  until False;

  if MultiComplexFlag then
    begin
      if ch = #0 then Abort ('Unexpected end of file!');
      if ch <> Ch_End_Compl then Abort ('End of complex not found.')
    end;
  DoneHash;
  if FacetCount = 0 then Abort ('No ' + Description + ' given!');
  PointCount := MaxIndex;
  ParseFacets := FacetList
end;

end.
